/*
 * Copyright (C) 2002 Roman Zippel <zippel@linux-m68k.org>
 * Released under the terms of the GNU GPL v2.0.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "expr.h"

struct file *file_list;
struct file *current_file;

struct symbol symbol_yes = {
	name: "y",
	curr: { "y", yes },
	flags: SYMBOL_YES|SYMBOL_DEP,
}, symbol_mod = {
	name: "m",
	curr: { "m", mod },
	flags: SYMBOL_MOD|SYMBOL_DEP,
}, symbol_no = {
	name: "n",
	curr: { "n", no },
	flags: SYMBOL_NO|SYMBOL_DEP,
}, symbol_empty = {
	name: "",
	curr: { "", no },
	flags: SYMBOL_DEP,
};

struct symbol *modules_sym;

struct expr *alloc_expr_symbol(struct symbol *sym)
{
	struct expr *e = malloc(sizeof(*e));
	memset(e, 0, sizeof(*e));
	e->token = T_WORD;
	e->left.sym = sym;
	return e;
}

struct expr *alloc_expr_one(int token, struct expr *ce)
{
	struct expr *e = malloc(sizeof(*e));
	memset(e, 0, sizeof(*e));
	e->token = token;
	e->left.expr = ce;
	return e;
}

struct expr *alloc_expr_two(int token, struct expr *e1, struct expr *e2)
{
	struct expr *e = malloc(sizeof(*e));
	memset(e, 0, sizeof(*e));
	e->token = token;
	e->left.expr = e1;
	e->right.expr = e2;
	return e;
}

struct expr *alloc_expr_comp(int token, struct symbol *s1, struct symbol *s2)
{
	struct expr *e = malloc(sizeof(*e));
	memset(e, 0, sizeof(*e));
	e->token = token;
	e->left.sym = s1;
	e->right.sym = s2;
	return e;
}

struct expr *copy_expr(struct expr *org)
{
	struct expr *e;

	if (!org)
		return NULL;

	e = malloc(sizeof(*org));
	memcpy(e, org, sizeof(*org));
	switch (org->token) {
	case T_WORD:
		e->left = org->left;
		break;
	case '!':
		e->left.expr = copy_expr(org->left.expr);
		break;
	case '=':
	case T_UNEQUAL:
		e->left.sym = org->left.sym;
		e->right.sym = org->right.sym;
		break;
	case T_OR:
	case T_AND:
	case '&':
	case '|':
	case '^':
	case T_IF:
	case T_ELSE:
		e->left.expr = copy_expr(org->left.expr);
		e->right.expr = copy_expr(org->right.expr);
		break;
	default:
		printf("can't copy token %d\n", e->token);
		free(e);
		e = NULL;
		break;
	}

	return e;
}

void free_expr(struct expr *e)
{
	if (!e)
		return;

	switch (e->token) {
	case T_WORD:
		break;
	case '!':
		free_expr(e->left.expr);
		return;
	case '=':
	case T_UNEQUAL:
		break;
	case '|':
	case T_OR:
	case T_AND:
	case '&':
	case T_IF:
	case T_ELSE:
		free_expr(e->left.expr);
		free_expr(e->right.expr);
		break;
	default:
		printf("how to free token %d?\n", e->token);
		break;
	}
	free(e);
}

tristate sym2expr(struct symbol *sym)
{
	if (sym->flags & SYMBOL_YES)
		return yes;
	else if (sym->flags & SYMBOL_MOD)
		return mod;
	return no;
}

struct symbol *expr2sym(tristate val)
{
	switch (val) {
	case yes:
		return &symbol_yes;
	case mod:
		return &symbol_mod;
	default:
		return &symbol_no;
	}
}

tristate calc_expr_value(struct symbol *sym, struct expr *e)
{
	int dep1, dep2;

	if (!e)
		return yes;

	switch (e->token) {
	case T_WORD:
		sym_calc_value(e->left.sym);
		//if ((e->left.sym->flags & SYMBOL_CHOICE) &&
		//    (S_TRI(e->left.sym->curr) == yes))
		//	return S_VAL(e->left.sym->curr) == sym ? yes : no;
		return S_TRI(e->left.sym->curr);
	case '&':
		dep1 = calc_expr_value(sym, e->left.expr);
		dep2 = calc_expr_value(sym, e->right.expr);
		return E_AND(dep1, dep2);
	case '|':
		dep1 = calc_expr_value(sym, e->left.expr);
		dep2 = calc_expr_value(sym, e->right.expr);
		return E_OR(dep1, dep2);
	case '!':
		dep1 = calc_expr_value(sym, e->left.expr);
		return E_NOT(dep1);
	case '=':
		sym_calc_value(e->left.sym);
		sym_calc_value(e->right.sym);
		return S_EQ(e->left.sym->curr, e->right.sym->curr) ? yes : no;
	case T_UNEQUAL:
		sym_calc_value(e->left.sym);
		sym_calc_value(e->right.sym);
		return S_EQ(e->left.sym->curr, e->right.sym->curr) ? no : yes;
	default:
		printf("calc_expr_value: %d?\n", e->token);
		return no;
	}
}

struct statement *sym_get_choice_stmt(struct symbol *sym)
{
	struct statement *stmt;

	for_all_choices(sym, stmt)
		return stmt;
	return NULL;
}

struct statement *sym_get_default_stmt(struct symbol *sym)
{
	struct statement *stmt;
	tristate visible;

	for_all_defaults(sym, stmt) {
		visible = E_CALC(sym, stmt->visible);
		if (visible != no)
			return stmt;
	}
	return NULL;
}

void sym_calc_visibility(struct symbol *sym)
{
	struct statement *stmt;
	tristate visible;

	/* any prompt visible? */
	visible = no;
	for_all_prompts(sym, stmt)
		visible = E_OR(visible, E_CALC(sym, stmt->visible));
	sym->visible = visible;
}

void sym_calc_value(struct symbol *sym)
{
	struct symbol_value newval, oldval;
	struct statement *stmt, *def_stmt;
	struct symbol *def_sym;
	struct expr *e;

	if (sym->flags & SYMBOL_DEP)
		return;

	//if (sym->name && !(strcmp(sym->name, "SD_EXTRA_DEVS")))
	//	printf("");

	oldval = sym->curr;

	switch (sym->type) {
	case SYMBOL_INT:
	case SYMBOL_HEX:
	case SYMBOL_STRING:
		newval = symbol_empty.curr;
		break;
	case SYMBOL_BOOLEAN:
	case SYMBOL_TRISTATE:
		newval = symbol_no.curr;
		break;
	default:
		S_VAL(newval) = sym->name;
		if (sym->flags & SYMBOL_CONST) {
			S_TRI(newval) = no;
			goto out;
		}
		//newval = symbol_empty.curr;
		// generate warning somewhere here later
		//S_TRI(newval) = yes;
		S_TRI(newval) = no;
		goto out;
	}
	sym->flags |= SYMBOL_DEP;
	if (!sym_is_choice_value(sym))
		sym->flags &= ~SYMBOL_WRITE;

	sym_calc_visibility(sym);

	/* set default if recursively called */
	sym->curr = newval;

	if (sym->visible != no) {
		sym->flags |= SYMBOL_WRITE;
		if (!sym_has_value(sym)) {
			stmt = sym_get_default_stmt(sym);
			if (stmt) {
				sym_calc_value(stmt->def);
				newval = stmt->def->curr;
			}
		} else
			newval = sym->def;

		S_TRI(newval) = E_AND(S_TRI(newval), sym->visible);
		/* if the symbol is visible and not optionial,
		 * possibly ignore old user choice. */
		if (!sym_is_optional(sym) && S_TRI(newval) == no)
			S_TRI(newval) = sym->visible;
		if (sym_is_choice_value(sym) && sym->visible == yes) {
			stmt = sym_get_choice_stmt(sym);
			S_TRI(newval) = (S_VAL(stmt->def->def) == sym) ? yes : no;
		}
	} else {
		stmt = sym_get_default_stmt(sym);
		if (stmt) {
			sym->flags |= SYMBOL_WRITE;
			sym_calc_value(stmt->def);
			newval = stmt->def->curr;
		}
	}

	switch (sym->type) {
	case SYMBOL_INT:
	case SYMBOL_HEX:
	case SYMBOL_STRING:
		// still possible???
		if (!S_VAL(newval))
			newval = symbol_empty.curr;
		goto out;
	}

	switch (sym_get_type(sym)) {
	case SYMBOL_TRISTATE:
		if (S_TRI(newval) != mod)
			break;
		sym_calc_value(modules_sym);
		if (S_TRI(modules_sym->curr) == no)
			S_TRI(newval) = yes;
		break;
	case SYMBOL_BOOLEAN:
		if (S_TRI(newval) == mod)
			S_TRI(newval) = yes;
	}

	// move this to sym_get_string_value?
	switch (S_TRI(newval)) {
	case no:
		S_VAL(newval) = S_VAL(symbol_no.curr);
		break;
	case mod:
		S_VAL(newval) = S_VAL(symbol_mod.curr);
		break;
	case yes:
		S_VAL(newval) = S_VAL(symbol_yes.curr);
		break;
	}
out:
	sym->curr = newval;

	if (sym_is_choice(sym) && S_TRI(newval) == yes) {
		def_sym = S_VAL(sym->def);
		if (def_sym) {
			sym_calc_visibility(def_sym);
			if (def_sym->visible == no)
				def_sym = NULL;
		}
		if (!def_sym) {
			for_all_defaults(sym, def_stmt) {
				if (E_CALC(sym, def_stmt->visible) == no)
					continue;
				sym_calc_visibility(def_stmt->def);
				if (def_stmt->def->visible != no) {
					def_sym = def_stmt->def;
					break;
				}
			}
		}

		if (!def_sym) {
			stmt = sym_get_choice_stmt(sym);
			for (e = stmt->dep; e; e = e->left.expr) {
				sym_calc_visibility(e->right.sym);
				if (e->right.sym->visible != no) {
					def_sym = e->right.sym;
					break;
				}
			}
		}

		S_VAL(newval) = def_sym;
	}

	if (memcmp(&oldval, &newval, sizeof(newval)))
		sym->flags |= SYMBOL_CHANGED;
	sym->curr = newval;

	if (sym_is_choice(sym)) {
		int flags = sym->flags & (SYMBOL_CHANGED | SYMBOL_WRITE);
		stmt = sym_get_choice_stmt(sym);
		for (e = stmt->dep; e; e = e->left.expr)
			e->right.sym->flags |= flags;
	}
}

void clear_sym_deps(void)
{
	struct symbol *sym;
	int i;

	for_all_symbols(i, sym)
		sym->flags &= ~SYMBOL_DEP;
}

bool sym_tristate_within_range(struct symbol *sym, tristate val)
{
	int type = sym_get_type(sym);

	if (sym->visible == no)
		return false;

	if (type != SYMBOL_BOOLEAN && type != SYMBOL_TRISTATE)
		return false;

	switch (val) {
	case no:
		if (sym_is_choice_value(sym) && sym->visible == yes)
			return false;
		return sym_is_optional(sym);
	case mod:
		if (sym_is_choice_value(sym) && sym->visible == yes)
			return false;
		return type == SYMBOL_TRISTATE;
	case yes:
		return type == SYMBOL_BOOLEAN || sym->visible == yes;
	}
	return false;
}

bool sym_set_tristate_value(struct symbol *sym, tristate val)
{
	if (sym_get_tristate_value(sym) != val &&
	    !sym_tristate_within_range(sym, val))
		return false;

	if (sym_is_choice_value(sym) && val == yes) {
		struct statement *stmt = sym_get_choice_stmt(sym);

		S_VAL(stmt->def->def) = sym;
		stmt->def->flags &= ~SYMBOL_NEW;
	} else
		S_TRI(sym->def) = val;

	sym->flags &= ~SYMBOL_NEW;
	if (sym_is_choice(sym)) {
		//if (!S_VAL(sym->def))
			// find default?
		//	S_VAL(sym->def) = sym;
		clear_sym_deps();
		sym_calc_value(sym);
		return true;
	}

	switch (val) {
	case no:
		S_VAL(sym->def) = S_VAL(symbol_no.curr);
		break;
	case mod:
		S_VAL(sym->def) = S_VAL(symbol_mod.curr);
		break;
	case yes:
		S_VAL(sym->def) = S_VAL(symbol_yes.curr);
		break;
	}
	return true;
}

tristate sym_toggle_tristate_value(struct symbol *sym)
{
	tristate oldval, newval;

	oldval = newval = sym_get_tristate_value(sym);
	do {
		switch (newval) {
		case no:
			newval = mod;
			break;
		case mod:
			newval = yes;
			break;
		case yes:
			newval = no;
			break;
		}
		if (sym_set_tristate_value(sym, newval))
			break;
	} while (oldval != newval);
	return newval;
}

bool sym_set_string_value(struct symbol *sym, const char *newval)
{
	free(S_VAL(sym->def));
	S_VAL(sym->def) = strdup(newval);
	S_VAL(sym->curr) = NULL;
	sym->flags &= ~SYMBOL_NEW;
	return true;
}

bool sym_is_changable(struct symbol *sym)
{
	if (sym->visible == no)
		return false;
	/* at least 'n' and 'y'/'m' is selectable */
	if (sym_is_optional(sym))
		return true;
	/* no 'n', so 'y' and 'm' must be selectable */
	if (sym_get_type(sym) == SYMBOL_TRISTATE && sym->visible == yes)
		return true;
	return false;
}

int contains_symbol(struct expr *dep, struct symbol *sym)
{
	if (!dep)
		return 0;

	switch (dep->token) {
	case T_AND:
	case T_OR:
	case '&':
	case '|':
		return contains_symbol(dep->left.expr, sym) ||
		       contains_symbol(dep->right.expr, sym);
	case T_WORD:
		return dep->left.sym == sym;
	case '=':
	case T_UNEQUAL:
		return dep->left.sym == sym;
	case '!':
		return contains_symbol(dep->left.expr, sym);
	}
	return 0;
}

bool depends_on_symbol(struct expr *dep, struct symbol *sym)
{
	if (!dep)
		return false;

	switch (dep->token) {
	case T_AND:
	case '&':
		return depends_on_symbol(dep->left.expr, sym) ||
		       depends_on_symbol(dep->right.expr, sym);
	case T_WORD:
		return dep->left.sym == sym;
	case '=':
		if (dep->left.sym == sym && (dep->right.sym == &symbol_yes || dep->right.sym == &symbol_mod))
			return true;
	case T_UNEQUAL:
		if (dep->left.sym == sym && dep->right.sym == &symbol_no)
			return true;
	}
 	return false;
}

static inline int token_gt(int t1, int t2)
{
#if 0
	return 1;
#else
	switch (t1) {
	case '=':
	case T_UNEQUAL:
		if (t2 == '!')
			return 1;
	case '!':
		if (t2 == T_AND || t2 == T_IF || t2 == T_ELSE || t2 == '&')
			return 1;
	case '&':
	case T_IF:
	case T_ELSE:
	case T_AND:
		if (t2 == T_OR || t2 == '|')
			return 1;
	case '|':
	case T_OR:
		if (t2 == '^')
			return 1;
	case '^':
		if (t2 == 0)
			return 1;
	case 0:
		return 0;
	}
	printf("[%dgt%d?]", t1, t2);
	return 0;
#endif
}

int print_type = 1;

static void print_sym(FILE *out, struct symbol *sym)
{
	fprintf(out, "%s", sym->name);
	if (print_type)
		fprintf(out, "%s",
			sym->type == SYMBOL_BOOLEAN ? "?" :
			sym->type == SYMBOL_TRISTATE ? "??" :
			sym->type == SYMBOL_HEX ? "h" :
			sym->type == SYMBOL_INT ? "i" :
			sym->type == SYMBOL_STRING ? "s" :
		       	"");
}

void fprint_expr(FILE *out, struct expr *e, int prevtoken)
{
	if (!e) {
		fprintf(out, "<none>");
		return;
	}

	switch (e->token) {
	case T_WORD:
		print_sym(out, e->left.sym);
		break;
	case '!':
		if (token_gt(prevtoken, '!'))
			fprintf(out, "(");
		fprintf(out, "!");
		fprint_expr(out, e->left.expr, '!');
		if (token_gt(prevtoken, '!'))
			fprintf(out, ")");
		break;
	case '=':
		if (token_gt(prevtoken, '='))
			fprintf(out, "(");
		print_sym(out, e->left.sym);
		fprintf(out, "=");
		print_sym(out, e->right.sym);
		if (token_gt(prevtoken, '='))
			fprintf(out, ")");
		break;
	case T_UNEQUAL:
		if (token_gt(prevtoken, T_UNEQUAL))
			fprintf(out, "(");
		print_sym(out, e->left.sym);
		fprintf(out, "!=");
		print_sym(out, e->right.sym);
		if (token_gt(prevtoken, T_UNEQUAL))
			fprintf(out, ")");
		break;
	case T_OR:
		if (token_gt(prevtoken, T_OR))
			fprintf(out, "(");
		fprint_expr(out, e->left.expr, T_OR);
		fprintf(out, " || ");
		fprint_expr(out, e->right.expr, T_OR);
		if (token_gt(prevtoken, T_OR))
			fprintf(out, ")");
		break;
	case T_AND:
		if (token_gt(prevtoken, T_AND))
			fprintf(out, "(");
		fprint_expr(out, e->left.expr, T_AND);
		fprintf(out, " && ");
		fprint_expr(out, e->right.expr, T_AND);
		if (token_gt(prevtoken, T_AND))
			fprintf(out, ")");
		break;
	case '|':
		if (token_gt(prevtoken, '|'))
			fprintf(out, "(");
		fprint_expr(out, e->left.expr, '|');
		fprintf(out, " || ");
		fprint_expr(out, e->right.expr, '|');
		if (token_gt(prevtoken, '|'))
			fprintf(out, ")");
		break;
	case '&':
		if (token_gt(prevtoken, '&'))
			fprintf(out, "(");
		fprint_expr(out, e->left.expr, '&');
		if (print_type)
			fprintf(out, " & ");
		else
			fprintf(out, " && ");
		fprint_expr(out, e->right.expr, '&');
		if (token_gt(prevtoken, '&'))
			fprintf(out, ")");
		break;
	case '^':
		if (e->left.expr) {
			fprint_expr(out, e->left.expr, '^');
			fprintf(out, " ^ ");
		}
		fprintf(out, "%s", e->right.sym->name);
		break;
	case T_IF:
		if (token_gt(prevtoken, T_IF))
			fprintf(out, "[");
		if (e->right.expr) {
			fprint_expr(out, e->right.expr, T_IF);
			fprintf(out, " && ");
		}
		fprint_expr(out, e->left.expr, T_IF);
		if (token_gt(prevtoken, T_IF))
			fprintf(out, "]");
		break;
	case T_ELSE:
		if (token_gt(prevtoken, T_ELSE))
			fprintf(out, "[");
		fprintf(out, "[");
		if (e->right.expr) {
			fprint_expr(out, e->right.expr, T_ELSE);
			fprintf(out, " && ");
		}
		fprintf(out, "!");
		fprint_expr(out, e->left.expr, '!');
		if (token_gt(prevtoken, T_ELSE))
			fprintf(out, "]");
		break;
	default:
		fprintf(out, "<unknown token %d>", e->token);
		break;
	}
}

void print_expr(int mask, struct expr *e, int prevtoken)
{
	if (!(cdebug & mask))
		return;
	fprint_expr(stdout, e, prevtoken);
}

struct file *lookup_file(const char *name)
{
	struct file *file;

	for (file = file_list; file; file = file->next) {
		if (!strcmp(name, file->name))
			return file;
	}

	file = malloc(sizeof(*file));
	memset(file, 0, sizeof(*file));
	file->name = strdup(name);
	file->next = file_list;
	file_list = file;
	return file;
}

bool menu_is_visible(struct menu *menu)
{
	tristate visible;
	if (menu->sym) {
		sym_calc_value(menu->sym);
		visible = menu->prompt ? E_TRI(menu->prompt->visible)
				       : no;
#if 0
		if (sym_is_choice(menu->sym)) {
			struct menu *cmenu;

			for (cmenu = menu->list; cmenu; cmenu = cmenu->next) {
				if (menu_is_visible(cmenu))
					break;
			}
			if (!cmenu)
				visible = no;
		}
#endif
	} else
		/* root menu is always visible, although it has no prompt */
		visible = menu->prompt ? E_CALC(NULL, menu->prompt->visible)
				       : yes;
	return visible != no;
}

char *menu_get_prompt(struct menu *menu)
{
	if (menu->prompt)
		return menu->prompt->prompt;
	else if (menu->sym)
		return menu->sym->name;
	return NULL;
}
